# Modified work Copyright (c) 2017-2021 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

# Original author: Pearu Peterson <pearu@cens.ioc.ee>
# First version created: May 2006

"""
Test parsing single Fortran lines.

"""

import pytest
from fparser.one.block_statements import (
    Allocatable,
    Allocate,
    ArithmeticIf,
    AssignedGoto,
    Assign,
    Assignment,
    Asynchronous,
    Backspace,
    Bind,
    Call,
    CallProtoArgument,
    CallStatement,
    Check,
    Close,
    Common,
    ComputedGoto,
    Contains,
    Continue,
    Cycle,
    Data,
    Deallocate,
    Depend,
    Dimension,
    Else,
    ElseIf,
    ElseWhere,
    Endfile,
    Entry,
    Enumerator,
    Equivalence,
    Exit,
    External,
    FinalBinding,
    Flush,
    Format,
    ForallStmt,
    FortranName,
    GenericBinding,
    Goto,
    Implicit,
    Import,
    Inquire,
    Intent,
    Intrinsic,
    ModuleProcedure,
    Nullify,
    Open,
    Optional,
    Parameter,
    Pointer,
    PointerAssignment,
    Pause,
    Print,
    Private,
    Protected,
    Public,
    Read,
    Return,
    Rewind,
    Save,
    Sequence,
    SpecificBinding,
    Stop,
    Target,
    Threadsafe,
    Use,
    Value,
    Volatile,
    Wait,
    WhereStmt,
    Write,
)
import fparser.common.sourceinfo
from fparser.one.typedecl_statements import (
    Byte,
    Character,
    Complex,
    DoubleComplex,
    DoublePrecision,
    Integer,
    Logical,
    Real,
)

from fparser.common.readfortran import FortranStringReader
from fparser.common.utils import AnalyzeError, ParseError


def parse(cls, line, label="", isfree=True, isstrict=False):
    """Tries to parse a Fortran line using the given class cls.
    If successful, it then converts the parsed statement back to
    a string. If isstrict is false, it will then try to parse
    this string again (recursively calling itself, with isstrict
    set to true) and make sure that the re-parsed string is
    identical to the input.
    It returns the string representation of the parsed input line.
    """

    if label:
        line = label + " : " + line
    reader = FortranStringReader(line)
    reader.set_format(fparser.common.sourceinfo.FortranFormat(isfree, isstrict))
    item = next(reader)
    if not cls.match(item.get_line()):
        raise ValueError("%r does not match %s pattern" % (line, cls.__name__))
    stmt = cls(item, item)
    if stmt.isvalid:
        # Check that we can successfully parse the string representation
        # of the parsed object
        stmt_string = str(stmt)
        if not isstrict:
            reparsed_stmt_string = parse(cls, stmt_string, isstrict=True)
            if stmt_string != reparsed_stmt_string:
                raise ValueError(
                    "Failed to parse %r with %s pattern in Pyf "
                    "mode, got %r" % (stmt_string, cls.__name__, reparsed_stmt_string)
                )
        return stmt_string
    raise ValueError("parsing %r with %s pattern failed" % (line, cls.__name__))


def test_assignment():
    """Test that assignment statements are parsed correctly."""
    assert parse(Assignment, "a=b") == "a = b"
    assert parse(PointerAssignment, "a=>b") == "a => b"
    assert parse(Assignment, "a (2)=b(n,m)") == "a(2) = b(n,m)"
    assert parse(Assignment, "a % 2(2,4)=b(a(i))") == "a%2(2,4) = b(a(i))"
    assert (
        parse(Assignment, "cu(i,j) = .5d0*(cu(i+1,j) + cu(i-1,j))")
        == "cu(i,j) = .5d0*(cu(i+1,j) + cu(i-1,j))"
    )


def test_assign():
    assert parse(Assign, "assign 10 to a") == "ASSIGN 10 TO a"


def test_call():
    assert parse(Call, "call a") == "CALL a"
    assert parse(Call, "call a()") == "CALL a"
    assert parse(Call, "call a(1)") == "CALL a(1)"
    assert parse(Call, "call a(1,2)") == "CALL a(1, 2)"
    assert parse(Call, "call a % c ( n , a+1 )") == "CALL a % c(n, a+1)"


def test_goto():
    assert parse(Goto, "go to 19") == "GO TO 19"
    assert parse(Goto, "goto 19") == "GO TO 19"
    assert parse(ComputedGoto, "goto (1, 2 ,3) a+b(2)") == "GO TO (1, 2, 3) a+b(2)"
    assert parse(ComputedGoto, "goto (1, 2 ,3) , a+b(2)") == "GO TO (1, 2, 3) a+b(2)"
    assert parse(AssignedGoto, "goto a") == "GO TO a"
    assert parse(AssignedGoto, "goto a ( 1 )") == "GO TO a (1)"
    assert parse(AssignedGoto, "goto a ( 1 ,2)") == "GO TO a (1, 2)"


def test_continue():
    assert parse(Continue, "continue") == "CONTINUE"


def test_return():
    assert parse(Return, "return") == "RETURN"
    assert parse(Return, "return a") == "RETURN a"
    assert parse(Return, "return a+1") == "RETURN a+1"
    assert parse(Return, "return a(c, a)") == "RETURN a(c, a)"


def test_stop():
    assert parse(Stop, "stop") == "STOP"
    assert parse(Stop, "stop 1") == "STOP 1"
    assert parse(Stop, 'stop "a"') == 'STOP "a"'
    assert parse(Stop, 'stop "a b"') == 'STOP "a b"'


def test_print():
    assert parse(Print, "print*") == "PRINT *"
    assert parse(Print, 'print "a b( c )"') == 'PRINT "a b( c )"'
    assert parse(Print, "print 12, a") == "PRINT 12, a"
    assert parse(Print, "print 12, a , b") == "PRINT 12, a, b"
    assert parse(Print, "print 12, a(c,1) , b") == "PRINT 12, a(c,1), b"


def test_read():
    assert parse(Read, "read ( 10 )") == "READ (10)"
    assert parse(Read, "read ( 10 ) a ") == "READ (10) a"
    assert parse(Read, "read ( 10 ) a , b") == "READ (10) a, b"
    assert parse(Read, "read *") == "READ *"
    assert parse(Read, "read 12") == "READ 12"
    assert parse(Read, 'read "a b"') == 'READ "a b"'
    assert parse(Read, 'read "a b",a') == 'READ "a b", a'
    assert parse(Read, "read * , a") == "READ *, a"
    assert parse(Read, 'read "hey a" , a') == 'READ "hey a", a'
    assert parse(Read, "read * , a  , b") == "READ *, a, b"
    assert parse(Read, "read ( unit  =10 )") == "READ (UNIT = 10)"


def test_write():
    assert parse(Write, "write ( 10 )") == "WRITE (10)"
    assert parse(Write, "write ( 10 , a )") == "WRITE (10, a)"
    assert parse(Write, "write ( 10 ) b") == "WRITE (10) b"
    assert parse(Write, "write ( 10 ) a(1) , b+2") == "WRITE (10) a(1), b+2"
    assert parse(Write, "write ( unit=10 )") == "WRITE (UNIT = 10)"


def test_flush():
    assert parse(Flush, "flush 10") == "FLUSH (10)"
    assert parse(Flush, "flush (10)") == "FLUSH (10)"
    assert parse(Flush, "flush (UNIT = 10)") == "FLUSH (UNIT = 10)"
    assert parse(Flush, "flush (10, err=  23)") == "FLUSH (10, ERR = 23)"


def test_wait():
    assert parse(Wait, "wait(10)") == "WAIT (10)"
    assert parse(Wait, "wait(10,err=129)") == "WAIT (10, ERR = 129)"


def test_contains():
    assert parse(Contains, "contains") == "CONTAINS"


def test_allocate():
    """Tests for various forms of ALLOCATE statement"""
    assert parse(Allocate, "allocate (a)") == "ALLOCATE (a)"
    assert parse(Allocate, "allocate (a, stat=b)") == "ALLOCATE (a, STAT = b)"
    assert parse(Allocate, "allocate (a,b(:1))") == "ALLOCATE (a, b(:1))"
    assert parse(Allocate, "allocate (real(8)::a)") == "ALLOCATE (REAL(KIND=8) :: a)"
    # With a type specifier
    assert parse(Allocate, "allocate (real :: a(8))") == "ALLOCATE (REAL :: a(8))"
    assert (
        parse(Allocate, "allocate (real(kind=wp) :: a(8))")
        == "ALLOCATE (REAL(KIND=wp) :: a(8))"
    )
    assert parse(Allocate, "allocate (a_type :: a)") == "ALLOCATE (a_type :: a)"
    with pytest.raises(ParseError) as err:
        parse(Allocate, "allocate(not valid :: a)")
    assert "Unrecognised type-specification in ALLOCATE statement" in str(err.value)


def test_deallocate():
    assert parse(Deallocate, "deallocate (a)") == "DEALLOCATE (a)"
    assert parse(Deallocate, "deallocate (a, stat=b)") == "DEALLOCATE (a, STAT = b)"


def test_moduleprocedure():
    """Test  [ MODULE ] PROCEDURE [::] <procedure-name-list>"""

    assert parse(ModuleProcedure, "Procedure a") == "MODULE PROCEDURE a"
    assert parse(ModuleProcedure, "procedure a , b") == "MODULE PROCEDURE a, b"
    assert parse(ModuleProcedure, "procedure :: a ") == "MODULE PROCEDURE a"
    assert parse(ModuleProcedure, "procedure :: a , b") == "MODULE PROCEDURE a, b"
    assert parse(ModuleProcedure, "ModuleProcedure a") == "MODULE PROCEDURE a"
    assert parse(ModuleProcedure, "module procedure a , b") == "MODULE PROCEDURE a, b"
    assert parse(ModuleProcedure, "module procedure :: a ") == "MODULE PROCEDURE a"
    assert (
        parse(ModuleProcedure, "module procedure :: a , b") == "MODULE PROCEDURE a, b"
    )
    assert parse(ModuleProcedure, "moduleprocedure::a,b") == "MODULE PROCEDURE a, b"


def test_access():
    assert parse(Public, "Public") == "PUBLIC"
    assert parse(Public, "public a") == "PUBLIC a"
    assert parse(Public, "public :: a") == "PUBLIC a"
    assert parse(Public, "public a,b,c") == "PUBLIC a, b, c"
    assert parse(Public, "public :: a(:,:)") == "PUBLIC a(:,:)"
    assert parse(Private, "private") == "PRIVATE"
    assert parse(Private, "private :: a") == "PRIVATE a"


def test_close():
    assert parse(Close, "close (12)") == "CLOSE (12)"
    assert parse(Close, "close (12, err=99)") == "CLOSE (12, ERR = 99)"
    assert parse(Close, "close (12, status = a(1,2))") == "CLOSE (12, STATUS = a(1,2))"


def test_class():
    """Check that we correctly parse and generate a class declaration"""
    from fparser.one.typedecl_statements import Class

    assert (
        parse(Class, "class(runtime_constants_type) :: a")
        == "CLASS(runtime_constants_type) a"
    )


def test_cycle():
    assert parse(Cycle, "cycle") == "CYCLE"
    assert parse(Cycle, "cycle ab") == "CYCLE ab"


def test_rewind():
    assert parse(Rewind, "rewind 1") == "REWIND (1)"
    assert parse(Rewind, "rewind (1)") == "REWIND (1)"
    assert parse(Rewind, "rewind (1, err =  123)") == "REWIND (1, ERR = 123)"


def test_backspace():
    assert parse(Backspace, "backspace 1") == "BACKSPACE (1)"
    assert parse(Backspace, "backspace (1)") == "BACKSPACE (1)"
    assert parse(Backspace, "backspace (1, err =  123)") == "BACKSPACE (1, ERR = 123)"


def test_endfile():
    assert parse(Endfile, "endfile 1") == "ENDFILE (1)"
    assert parse(Endfile, "endfile (1)") == "ENDFILE (1)"
    assert parse(Endfile, "endfile (1, err =  123)") == "ENDFILE (1, ERR = 123)"


def test_open():
    assert parse(Open, "open (1)") == "OPEN (1)"
    assert parse(Open, "open (1, err =  123)") == "OPEN (1, ERR = 123)"


def test_format():
    assert parse(Format, "1 format ()") == "1 FORMAT ()"
    assert parse(Format, "199 format (1)") == "199 FORMAT (1)"
    assert parse(Format, "2 format (1 , SS)") == "2 FORMAT (1, ss)"


def test_save():
    assert parse(Save, "save") == "SAVE"
    assert parse(Save, "save :: a") == "SAVE a"
    assert parse(Save, "save a,b") == "SAVE a, b"


def test_data():
    assert parse(Data, "data a /b/") == "DATA a / b /"
    assert parse(Data, "data a , c /b/") == "DATA a, c / b /"
    assert parse(Data, "data a /b ,c/") == "DATA a / b, c /"
    assert parse(Data, "data a /b/ c,e /d/") == "DATA a / b / c, e / d /"
    assert parse(Data, "data a(1,2) /b/") == "DATA a(1,2) / b /"
    assert parse(Data, "data a /b, c(1)/") == "DATA a / b, c(1) /"


def test_nullify():
    assert parse(Nullify, "nullify(a)") == "NULLIFY (a)"
    assert parse(Nullify, "nullify(a  ,b)") == "NULLIFY (a, b)"


def test_use():
    """Check that we can parse USE statement."""
    assert parse(Use, "use a") == "USE a"
    assert parse(Use, "use :: a") == "USE a"
    assert parse(Use, "use, intrinsic:: a") == "USE, INTRINSIC :: a"
    assert parse(Use, "use, non_intrinsic:: a") == "USE, NON_INTRINSIC :: a"
    assert (
        parse(Use, "use, non_intrinsic:: a , only: b")
        == "USE, NON_INTRINSIC :: a, ONLY: b"
    )
    assert parse(Use, "use :: a ,only: b") == "USE a, ONLY: b"
    assert parse(Use, "use :: a , only: b=>c") == "USE a, ONLY: b=>c"
    assert parse(Use, "use :: a , b=>c") == "USE a, b=>c"
    assert (
        parse(Use, "use :: a , only: operator(+) , b") == "USE a, ONLY: operator(+), b"
    )


def test_exit():
    assert parse(Exit, "exit") == "EXIT"
    assert parse(Exit, "exit ab") == "EXIT ab"


def test_parameter():
    assert parse(Parameter, "parameter (a = b(1,2))") == "PARAMETER (a = b(1,2))"
    assert (
        parse(Parameter, "parameter (a = b(1,2) , b=1)")
        == "PARAMETER (a = b(1,2), b=1)"
    )


def test_equivalence():
    assert parse(Equivalence, "equivalence (a , b)") == "EQUIVALENCE (a, b)"
    assert (
        parse(Equivalence, "equivalence (a , b) , ( c, d(1) , g  )")
        == "EQUIVALENCE (a, b), (c, d(1), g)"
    )


def test_dimension():
    assert parse(Dimension, "dimension a(b)") == "DIMENSION a(b)"
    assert parse(Dimension, "dimension::a(b)") == "DIMENSION a(b)"
    assert parse(Dimension, "dimension a(b)  , c(d)") == "DIMENSION a(b), c(d)"
    assert parse(Dimension, "dimension a(b,c)") == "DIMENSION a(b,c)"


def test_target():
    assert parse(Target, "target a(b)") == "TARGET a(b)"
    assert parse(Target, "target::a(b)") == "TARGET a(b)"
    assert parse(Target, "target a(b)  , c(d)") == "TARGET a(b), c(d)"
    assert parse(Target, "target a(b,c)") == "TARGET a(b,c)"


def test_pointer():
    assert parse(Pointer, "pointer a=b") == "POINTER a=b"
    assert parse(Pointer, "pointer :: a=b") == "POINTER a=b"
    assert parse(Pointer, "pointer a=b, c=d(1,2)") == "POINTER a=b, c=d(1,2)"


def test_protected():
    assert parse(Protected, "protected a") == "PROTECTED a"
    assert parse(Protected, "protected::a") == "PROTECTED a"
    assert parse(Protected, "protected a , b") == "PROTECTED a, b"


def test_volatile():
    assert parse(Volatile, "volatile a") == "VOLATILE a"
    assert parse(Volatile, "volatile::a") == "VOLATILE a"
    assert parse(Volatile, "volatile a , b") == "VOLATILE a, b"


def test_value():
    assert parse(Value, "value a") == "VALUE a"
    assert parse(Value, "value::a") == "VALUE a"
    assert parse(Value, "value a , b") == "VALUE a, b"


def test_arithmeticif():
    assert parse(ArithmeticIf, "if (a) 1,2,3") == "IF (a) 1, 2, 3"
    assert parse(ArithmeticIf, "if (a(1)) 1,2,3") == "IF (a(1)) 1, 2, 3"
    assert parse(ArithmeticIf, "if (a(1,2)) 1,2,3") == "IF (a(1,2)) 1, 2, 3"


def test_intrinsic():
    assert parse(Intrinsic, "intrinsic a") == "INTRINSIC a"
    assert parse(Intrinsic, "intrinsic::a") == "INTRINSIC a"
    assert parse(Intrinsic, "intrinsic a , b") == "INTRINSIC a, b"


def test_inquire():
    assert parse(Inquire, "inquire (1)") == "INQUIRE (1)"
    assert parse(Inquire, "inquire (1, err=123)") == "INQUIRE (1, ERR = 123)"
    assert parse(Inquire, "inquire (iolength=a) b") == "INQUIRE (IOLENGTH = a) b"
    assert (
        parse(Inquire, "inquire (iolength=a) b  ,c(1,2)")
        == "INQUIRE (IOLENGTH = a) b, c(1,2)"
    )


def test_sequence():
    assert parse(Sequence, "sequence") == "SEQUENCE"


def test_external():
    assert parse(External, "external a") == "EXTERNAL a"
    assert parse(External, "external::a") == "EXTERNAL a"
    assert parse(External, "external a , b") == "EXTERNAL a, b"


def test_common():
    assert parse(Common, "common a") == "COMMON a"
    assert parse(Common, "common a , b") == "COMMON a, b"
    assert parse(Common, "common a , b(1,2)") == "COMMON a, b(1,2)"
    assert parse(Common, "common // a") == "COMMON a"
    assert parse(Common, "common / name/ a") == "COMMON / name / a"
    assert parse(Common, "common / name/ a  , c") == "COMMON / name / a, c"
    assert (
        parse(Common, "common / name/ a /foo/ c(1) ,d")
        == "COMMON / name / a / foo / c(1), d"
    )
    assert (
        parse(Common, "common / name/ a, /foo/ c(1) ,d")
        == "COMMON / name / a / foo / c(1), d"
    )


def test_optional():
    assert parse(Optional, "optional a") == "OPTIONAL a"
    assert parse(Optional, "optional::a") == "OPTIONAL a"
    assert parse(Optional, "optional a , b") == "OPTIONAL a, b"


def test_intent():
    assert parse(Intent, "intent (in) a") == "INTENT (IN) a"
    assert parse(Intent, "intent(in)::a") == "INTENT (IN) a"
    assert parse(Intent, "intent(in) a , b") == "INTENT (IN) a, b"
    assert parse(Intent, "intent (in, out) a") == "INTENT (IN, OUT) a"


def test_entry():
    assert parse(Entry, "entry a") == "ENTRY a"
    assert parse(Entry, "entry a()") == "ENTRY a"
    assert parse(Entry, "entry a(b)") == "ENTRY a (b)"
    assert parse(Entry, "entry a(b,*)") == "ENTRY a (b, *)"
    assert (
        parse(Entry, 'entry a bind(c , name="a b")') == 'ENTRY a BIND (C, NAME = "a b")'
    )
    assert parse(Entry, "entry a result (b)") == "ENTRY a RESULT (b)"
    assert parse(Entry, "entry a bind(d) result (b)") == "ENTRY a RESULT (b) BIND (D)"
    assert parse(Entry, "entry a result (b) bind( c )") == "ENTRY a RESULT (b) BIND (C)"
    assert parse(Entry, "entry a(b,*) result (g)") == "ENTRY a (b, *) RESULT (g)"


def test_import():
    assert parse(Import, "import") == "IMPORT"
    assert parse(Import, "import a") == "IMPORT a"
    assert parse(Import, "import::a") == "IMPORT a"
    assert parse(Import, "import a , b") == "IMPORT a, b"


def test_forall():
    assert (
        parse(ForallStmt, "forall (i = 1:n(k,:) : 2) a(i) = i*i*b(i)")
        == "FORALL (i = 1 : n(k,:) : 2) a(i) = i*i*b(i)"
    )
    assert (
        parse(ForallStmt, "forall (i=1:n,j=2:3) a(i) = b(i,i)")
        == "FORALL (i = 1 : n, j = 2 : 3) a(i) = b(i,i)"
    )
    assert (
        parse(ForallStmt, "forall (i=1:n,j=2:3, 1+a(1,2)) a(i) = b(i,i)")
        == "FORALL (i = 1 : n, j = 2 : 3, 1+a(1,2)) a(i) = b(i,i)"
    )


def test_specificbinding():
    assert parse(SpecificBinding, "procedure a") == "PROCEDURE a"
    assert parse(SpecificBinding, "procedure :: a") == "PROCEDURE a"
    assert (
        parse(SpecificBinding, "procedure , NOPASS :: a") == "PROCEDURE , NOPASS :: a"
    )
    assert (
        parse(SpecificBinding, "procedure , public, pass(x ) :: a")
        == "PROCEDURE , PUBLIC, PASS (x) :: a"
    )
    assert parse(SpecificBinding, "procedure(n) a") == "PROCEDURE (n) a"
    assert (
        parse(SpecificBinding, "procedure(n), pass :: a") == "PROCEDURE (n) , PASS :: a"
    )
    assert parse(SpecificBinding, "procedure(n) :: a") == "PROCEDURE (n) a"
    assert parse(SpecificBinding, "procedure a= >b") == "PROCEDURE a => b"
    assert (
        parse(SpecificBinding, "procedure(n),pass :: a =>c")
        == "PROCEDURE (n) , PASS :: a => c"
    )


def test_genericbinding():
    assert parse(GenericBinding, "generic :: a=>b") == "GENERIC :: a => b"
    assert (
        parse(GenericBinding, "generic, public :: a=>b") == "GENERIC, PUBLIC :: a => b"
    )
    assert (
        parse(GenericBinding, "generic, public :: a(1,2)=>b ,c")
        == "GENERIC, PUBLIC :: a(1,2) => b, c"
    )


def test_finalbinding():
    assert parse(FinalBinding, "final a") == "FINAL a"
    assert parse(FinalBinding, "final::a") == "FINAL a"
    assert parse(FinalBinding, "final a , b") == "FINAL a, b"


def test_allocatable():
    assert parse(Allocatable, "allocatable a") == "ALLOCATABLE a"
    assert parse(Allocatable, "allocatable :: a") == "ALLOCATABLE a"
    assert parse(Allocatable, "allocatable a (1,2)") == "ALLOCATABLE a (1,2)"
    assert parse(Allocatable, "allocatable a (1,2) ,b") == "ALLOCATABLE a (1,2), b"


def test_asynchronous():
    assert parse(Asynchronous, "asynchronous a") == "ASYNCHRONOUS a"
    assert parse(Asynchronous, "asynchronous::a") == "ASYNCHRONOUS a"
    assert parse(Asynchronous, "asynchronous a , b") == "ASYNCHRONOUS a, b"


def test_bind():
    assert parse(Bind, "bind(c) a") == "BIND (C) a"
    assert parse(Bind, "bind(c) :: a") == "BIND (C) a"
    assert parse(Bind, "bind(c) a ,b") == "BIND (C) a, b"
    assert parse(Bind, "bind(c) /a/") == "BIND (C) / a /"
    assert parse(Bind, "bind(c) /a/ ,b") == "BIND (C) / a /, b"
    assert parse(Bind, 'bind(c,name="hey") a') == 'BIND (C, NAME = "hey") a'


def test_else():
    assert parse(Else, "else") == "ELSE"
    assert parse(ElseIf, "else if (a) then") == "ELSE IF (a) THEN"
    assert parse(ElseIf, "else if (a.eq.b(1,2)) then") == "ELSE IF (a.eq.b(1,2)) THEN"


def test_where():
    assert parse(WhereStmt, "where (1) a=1") == "WHERE ( 1 ) a = 1"
    assert parse(WhereStmt, "where (a(1,2)) a=1") == "WHERE ( a(1,2) ) a = 1"


def test_elsewhere():
    assert parse(ElseWhere, "else where") == "ELSE WHERE"
    assert parse(ElseWhere, "elsewhere (1)") == "ELSE WHERE ( 1 )"
    assert parse(ElseWhere, "elsewhere(a(1,2))") == "ELSE WHERE ( a(1,2) )"


def test_enumerator():
    assert parse(Enumerator, "enumerator a") == "ENUMERATOR a"
    assert parse(Enumerator, "enumerator:: a") == "ENUMERATOR a"
    assert parse(Enumerator, "enumerator a,b") == "ENUMERATOR a, b"
    assert parse(Enumerator, "enumerator a=1") == "ENUMERATOR a=1"
    assert parse(Enumerator, "enumerator a=1 , b=c(1,2)") == "ENUMERATOR a=1, b=c(1,2)"


def test_fortranname():
    assert parse(FortranName, "fortranname a") == "FORTRANNAME a"


def test_threadsafe():
    assert parse(Threadsafe, "threadsafe") == "THREADSAFE"


def test_depend():
    assert parse(Depend, "depend( a) b") == "DEPEND ( a ) b"
    assert parse(Depend, "depend( a) ::b") == "DEPEND ( a ) b"
    assert parse(Depend, "depend( a,c) b,e") == "DEPEND ( a, c ) b, e"


def test_check():
    assert parse(Check, "check(1) a") == "CHECK ( 1 ) a"
    assert parse(Check, "check(1) :: a") == "CHECK ( 1 ) a"
    assert parse(Check, "check(b(1,2)) a") == "CHECK ( b(1,2) ) a"
    assert parse(Check, "check(a>1) :: a") == "CHECK ( a>1 ) a"


def test_callstatement():
    assert (
        parse(CallStatement, "callstatement (*func)()", isstrict=1)
        == "CALLSTATEMENT (*func)()"
    )
    assert (
        parse(CallStatement, "callstatement i=1;(*func)()", isstrict=1)
        == "CALLSTATEMENT i=1;(*func)()"
    )


def test_callprotoargument():
    assert (
        parse(CallProtoArgument, "callprotoargument int(*), double")
        == "CALLPROTOARGUMENT int(*), double"
    )


def test_pause():
    assert parse(Pause, "pause") == "PAUSE"
    assert parse(Pause, "pause 1") == "PAUSE 1"
    assert parse(Pause, 'pause "hey"') == 'PAUSE "hey"'
    assert parse(Pause, 'pause "hey pa"') == 'PAUSE "hey pa"'


def test_byte():
    """Tests various declarations of byte variables."""
    assert parse(Byte, "byte") == "BYTE"
    assert parse(Byte, "byte a") == "BYTE a"
    assert parse(Byte, "byte, a") == "BYTE a"
    assert parse(Byte, "byte a ,b") == "BYTE a, b"
    assert parse(Byte, "byte :: a ,b") == "BYTE a, b"
    assert parse(Byte, "byte a(1,2)") == "BYTE a(1,2)"
    assert parse(Byte, "byte :: a(1,2),b") == "BYTE a(1,2), b"
    assert parse(Byte, "byte external :: a") == "BYTE, external :: a"
    assert parse(Byte, "byte, external :: a") == "BYTE, external :: a"
    assert (
        parse(Byte, "byte external , intent(in) :: a")
        == "BYTE, external, intent(in) :: a"
    )


def test_integer():
    """Tests various declarations of integer variables."""
    assert parse(Integer, "integer") == "INTEGER"
    assert parse(Integer, "integer*4") == "INTEGER*4"
    assert parse(Integer, "integer*4 a") == "INTEGER*4 a"
    assert parse(Integer, "integer*4, a") == "INTEGER*4 a"
    assert parse(Integer, "integer*4 a ,b") == "INTEGER*4 a, b"
    assert parse(Integer, "integer*4 :: a ,b") == "INTEGER*4 a, b"
    assert parse(Integer, "integer*4 a(1,2)") == "INTEGER*4 a(1,2)"
    assert parse(Integer, "integer*4 a(2*(1+2)-1)") == "INTEGER*4 a(2*(1+2)-1)"
    assert parse(Integer, "integer*4 :: a(1,2),b") == "INTEGER*4 a(1,2), b"
    assert parse(Integer, "integer*4 external :: a") == "INTEGER*4, external :: a"
    assert parse(Integer, "integer*4, external :: a") == "INTEGER*4, external :: a"
    assert (
        parse(Integer, "integer*4 external , intent(in) :: a")
        == "INTEGER*4, external, intent(in) :: a"
    )
    assert parse(Integer, "integer(kind=4)") == "INTEGER(KIND=4)"
    assert parse(Integer, "integer ( kind = 4)") == "INTEGER(KIND=4)"
    assert parse(Integer, "integer(kind=2+2)") == "INTEGER(KIND=2+2)"
    assert parse(Integer, "integer(kind=f(4,5))") == "INTEGER(KIND=f(4,5))"
    assert parse(Integer, "integer a = 5") == "INTEGER :: a = 5"


def test_const_array_decl():
    """Tests declarations that set an initial value to an array."""
    assert parse(Integer, "integer a(1) = (/1/)") == "INTEGER :: a(1) = (/1/)"
    assert (
        parse(Real, "real a(3) = (/1.1, 2.2, 3.3/)")
        == "REAL :: a(3) = (/1.1, 2.2, 3.3/)"
    )
    assert (
        parse(Logical, "logical :: a(3) = (/true, false, true/)")
        == "LOGICAL :: a(3) = (/true, false, true/)"
    )
    assert parse(Integer, "integer a(1) = [1]") == "INTEGER :: a(1) = [1]"
    assert parse(Integer, "integer a(3) = [1, 2, 3]") == "INTEGER :: a(3) = [1, 2, 3]"


def test_logical():
    """Tests various declarations of logical variables."""
    assert parse(Logical, "logical") == "LOGICAL"
    assert parse(Logical, "logical*4") == "LOGICAL*4"
    assert parse(Logical, "logical*4 a") == "LOGICAL*4 a"
    assert parse(Logical, "logical*4, a") == "LOGICAL*4 a"
    assert parse(Logical, "logical*4 a ,b") == "LOGICAL*4 a, b"
    assert parse(Logical, "logical*4 :: a ,b") == "LOGICAL*4 a, b"
    assert parse(Logical, "logical*4 a(1,2)") == "LOGICAL*4 a(1,2)"
    assert parse(Logical, "logical*4 :: a(1,2),b") == "LOGICAL*4 a(1,2), b"
    assert parse(Logical, "logical*4 external :: a") == "LOGICAL*4, external :: a"
    assert parse(Logical, "logical*4, external :: a") == "LOGICAL*4, external :: a"
    assert (
        parse(Logical, "logical*4 external , intent(in) :: a")
        == "LOGICAL*4, external, intent(in) :: a"
    )
    assert parse(Logical, "logical(kind=4)") == "LOGICAL(KIND=4)"
    assert parse(Logical, "logical ( kind = 4)") == "LOGICAL(KIND=4)"
    assert parse(Logical, "logical(kind=2+2)") == "LOGICAL(KIND=2+2)"
    assert parse(Logical, "logical(kind=f(4,5))") == "LOGICAL(KIND=f(4,5))"


def test_real():
    """Test various declarations of real variables."""

    assert parse(Real, "real") == "REAL"
    assert parse(Real, "real*4") == "REAL*4"
    assert parse(Real, "real*4 a") == "REAL*4 a"
    assert parse(Real, "real*4, a") == "REAL*4 a"
    assert parse(Real, "real*4 a ,b") == "REAL*4 a, b"
    assert parse(Real, "real*4 :: a ,b") == "REAL*4 a, b"
    assert parse(Real, "real*4 a(1,2)") == "REAL*4 a(1,2)"
    assert parse(Real, "real*4 :: a(1,2),b") == "REAL*4 a(1,2), b"
    assert parse(Real, "real*4 external :: a") == "REAL*4, external :: a"
    assert parse(Real, "real*4, external :: a") == "REAL*4, external :: a"
    assert (
        parse(Real, "real*4 external , intent(in) :: a")
        == "REAL*4, external, intent(in) :: a"
    )
    assert parse(Real, "real(kind=4)") == "REAL(KIND=4)"
    assert parse(Real, "real ( kind = 4)") == "REAL(KIND=4)"
    assert parse(Real, "real(kind=2+2)") == "REAL(KIND=2+2)"
    assert parse(Real, "real(kind=f(4,5))") == "REAL(KIND=f(4,5))"


def test_double_precision():
    """Test various declarations of double precision variables."""

    assert parse(DoublePrecision, "doubleprecision") == "DOUBLEPRECISION"
    assert parse(DoublePrecision, "double precision") == "DOUBLEPRECISION"
    assert parse(DoublePrecision, "doubleprecision a") == "DOUBLEPRECISION a"
    assert parse(DoublePrecision, "double precision, a") == "DOUBLEPRECISION a"
    assert parse(DoublePrecision, "doubleprecision a ,b") == "DOUBLEPRECISION a, b"
    assert parse(DoublePrecision, "double precision :: a ,b") == "DOUBLEPRECISION a, b"
    assert parse(DoublePrecision, "doubleprecision a(1,2)") == "DOUBLEPRECISION a(1,2)"
    assert (
        parse(DoublePrecision, "double precision :: a(1,2),b")
        == "DOUBLEPRECISION a(1,2), b"
    )
    assert (
        parse(DoublePrecision, "doubleprecision external :: a")
        == "DOUBLEPRECISION, external :: a"
    )
    assert (
        parse(DoublePrecision, "double precision, external :: a")
        == "DOUBLEPRECISION, external :: a"
    )
    assert (
        parse(DoublePrecision, "double precision external , intent(in) :: a")
        == "DOUBLEPRECISION, external, intent(in) :: a"
    )


def test_complex():
    """Test various declarations of complex variables."""

    assert parse(Complex, "complex") == "COMPLEX"
    assert parse(Complex, "complex") == "COMPLEX"
    assert parse(Complex, "complex*8 a") == "COMPLEX*8 a"
    assert parse(Complex, "complex*16, a") == "COMPLEX*16 a"
    assert parse(Complex, "complex a ,b") == "COMPLEX a, b"
    assert parse(Complex, "complex :: a ,b") == "COMPLEX a, b"
    assert parse(Complex, "complex a(1,2)") == "COMPLEX a(1,2)"
    assert parse(Complex, "complex :: a(1,2),b") == "COMPLEX a(1,2), b"
    assert parse(Complex, "complex external :: a") == "COMPLEX, external :: a"
    assert parse(Complex, "complex, external :: a") == "COMPLEX, external :: a"
    assert (
        parse(Complex, "complex external , intent(in) :: a")
        == "COMPLEX, external, intent(in) :: a"
    )
    assert parse(Complex, "complex(kind=8)") == "COMPLEX(KIND=8)"
    assert parse(Complex, "complex ( kind = 16)") == "COMPLEX(KIND=16)"
    assert parse(Complex, "complex(kind=4+4)") == "COMPLEX(KIND=4+4)"
    assert parse(Complex, "complex(kind=f(4,5))") == "COMPLEX(KIND=f(4,5))"


def test_double_complex():
    """Test various declarations of double complex variables."""

    assert parse(DoubleComplex, "double complex") == "DOUBLECOMPLEX"
    assert parse(DoubleComplex, "double complex") == "DOUBLECOMPLEX"
    assert parse(DoubleComplex, "double complex a") == "DOUBLECOMPLEX a"
    assert parse(DoubleComplex, "double complex, a") == "DOUBLECOMPLEX a"
    assert parse(DoubleComplex, "double complex a ,b") == "DOUBLECOMPLEX a, b"
    assert parse(DoubleComplex, "double complex :: a ,b") == "DOUBLECOMPLEX a, b"
    assert parse(DoubleComplex, "double complex a(1,2)") == "DOUBLECOMPLEX a(1,2)"
    assert (
        parse(DoubleComplex, "double complex :: a(1,2),b") == "DOUBLECOMPLEX a(1,2), b"
    )
    assert (
        parse(DoubleComplex, "double complex external :: a")
        == "DOUBLECOMPLEX, external :: a"
    )
    assert (
        parse(DoubleComplex, "double complex, external :: a")
        == "DOUBLECOMPLEX, external :: a"
    )
    assert (
        parse(DoubleComplex, "double complex external , intent(in) :: a")
        == "DOUBLECOMPLEX, external, intent(in) :: a"
    )
    assert parse(DoubleComplex, "double complex(kind=4)") == "DOUBLECOMPLEX(KIND=4)"
    assert parse(DoubleComplex, "double complex ( kind = 4)") == "DOUBLECOMPLEX(KIND=4)"
    assert parse(DoubleComplex, "double complex(kind=2+2)") == "DOUBLECOMPLEX(KIND=2+2)"
    assert (
        parse(DoubleComplex, "double complex(kind=f(4,5))")
        == "DOUBLECOMPLEX(KIND=f(4,5))"
    )


def test_character():
    assert parse(Character, "character") == "CHARACTER"
    assert parse(Character, "character*2") == "CHARACTER(LEN=2)"
    assert parse(Character, "character**") == "CHARACTER(LEN=*)"
    assert parse(Character, "character*(2)") == "CHARACTER(LEN=2)"
    assert parse(Character, "character*(len =2)") == "CHARACTER(LEN=2)"
    assert parse(Character, "character*(len =2)") == "CHARACTER(LEN=2)"
    assert parse(Character, "character*(len =:)") == "CHARACTER(LEN=:)"
    assert parse(Character, "character(len =2)") == "CHARACTER(LEN=2)"
    assert parse(Character, "character(2)") == "CHARACTER(LEN=2)"
    assert parse(Character, "character(kind=2)") == "CHARACTER(KIND=2)"
    assert parse(Character, "character(kind=2,len=3)") == "CHARACTER(LEN=3, KIND=2)"
    assert parse(Character, "character(lEN=3,kind=2)") == "CHARACTER(LEN=3, KIND=2)"
    assert (
        parse(Character, "character(len=3,kind=2)", isstrict=True)
        == "CHARACTER(LEN=3, KIND=2)"
    )
    assert (
        parse(Character, "chaRACTER(len=3,kind=fA(1,2))", isstrict=True)
        == "CHARACTER(LEN=3, KIND=fA(1,2))"
    )
    assert (
        parse(Character, "character(len=3,kind=fA(1,2))")
        == "CHARACTER(LEN=3, KIND=fa(1,2))"
    )


def test_implicit():
    assert parse(Implicit, "implicit none") == "IMPLICIT NONE"
    assert parse(Implicit, "implicit") == "IMPLICIT NONE"
    assert parse(Implicit, "implicit integer (i-m)") == "IMPLICIT INTEGER ( i-m )"
    assert (
        parse(Implicit, "implicit integer (i-m,p,q-r)")
        == "IMPLICIT INTEGER ( i-m, p, q-r )"
    )
    assert (
        parse(Implicit, "implicit integer (i-m), real (z)")
        == "IMPLICIT INTEGER ( i-m ), REAL ( z )"
    )


def test_type_bound_array_access():
    """Check that we can parse code that calls a type-bound procedure
    on the element of an array (of derived types)"""
    parsed_code = parse(Call, "call an_array(idx)%a_func(arg)")
    assert parsed_code == "CALL an_array(idx)%a_func(arg)"
    # Routine being called has no arguments - it is valid Fortran to
    # omit the parentheses entirely in this case
    parsed_code = parse(Call, "call an_array(idx)%a_func()")
    assert parsed_code == "CALL an_array(idx)%a_func"
    parsed_code = parse(Call, "call an_array(idx)%a_func")
    assert parsed_code == "CALL an_array(idx)%a_func"
    # Perversely put in a character string arg containing '('
    parsed_code = parse(Call, 'call an_array(idx)%a_func("(")')
    assert parsed_code == 'CALL an_array(idx)%a_func("(")'
    # Pass an array element as an argument
    parsed_code = parse(Call, "call an_array(idx)%a_func(b(3))")
    assert parsed_code == "CALL an_array(idx)%a_func(b(3))"


def test_invalid_type_bound_array_access():  # pylint: disable=invalid-name
    """Check that a call to a type-bound procedure with incorrect
    syntax is flagged as invalid"""
    with pytest.raises(ValueError) as excinfo:
        _ = parse(Call, "call an_array(idx)%a_func)")
    assert "with Call pattern failed" in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        _ = parse(Call, "call an_array(idx)%)")
    assert "with Call pattern failed" in str(excinfo.value)


def test_analyze_errors():
    """Tests that AnalyzeErrors are raised as expected. It also tests
    for various calling-sequence issues, e.g. parsing or analyzing twice,
    or calling analyze() without calling parse() first."""
    from fparser import api

    source_str = """none subroutine test()
      end
    """
    with pytest.raises(AnalyzeError) as error:
        _ = api.parse(source_str, isfree=True, isstrict=False)
    assert "no parse pattern found" in str(error.value)

    source_str = """subroutine test()
        incorrect :: c
      end
    """
    with pytest.raises(AnalyzeError) as error:
        _ = api.parse(source_str, isfree=False, isstrict=False)
    assert "no parse pattern found" in str(error.value)

    source_str = """subroutine test()
      end
    """
    from fparser.one.parsefortran import FortranParser

    reader = api.get_reader(source_str)
    parser = FortranParser(reader)

    # Handle analyze before parsing (does not raise an exception atm)
    parser.analyze()
    # Cover case that parsing is called twice
    parser.parse()
    parser.parse()

    # Check if analyse is called twice
    parser.analyze()
    parser.analyze()



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Parent_Identifier rule R1118
"""

from fparser.two.utils import Base


class Parent_Identifier(Base):  # R1118 (C1113)
    """Fortran 2008 rule R1118
    parent-identifier is ancestor-module-name [ : parent-submodule-name ]

    C1113 The ancestor-module-name shall be the name of a nonintrinsic
    module; the parent-submodule name shall be the name of a
    descendant of that module.
    This constraint can not be tested by fparser in general as the
    module or submodule may be in a different file. We therefore do
    not check this constraint in fparser.

    """

    use_names = ["Ancestor_Module_Name", "Parent_SubModule_Name"]

    @staticmethod
    def match(fstring):
        """Check whether the input matches the rule

        param string fstring : contains the Fortran that we are trying
        to match

        :return: instances of the Classes that have matched if there
        is a match or `None` if there is no match

        """
        # Avoid circular dependencies by importing here.
        # pylint: disable=import-outside-toplevel
        from fparser.two.Fortran2008 import Ancestor_Module_Name, Parent_SubModule_Name

        split_string = fstring.split(":")
        len_split_string = len(split_string)
        lhs_name = split_string[0].lstrip().rstrip()
        if len_split_string == 1:
            return Ancestor_Module_Name(lhs_name), None
        if len_split_string == 2:
            rhs_name = split_string[1].lstrip().rstrip()
            return Ancestor_Module_Name(lhs_name), Parent_SubModule_Name(rhs_name)
        # we expect at most one ':' in our input so the match fails
        return None

    def tostr(self):
        """return the fortran representation of this object"""
        # return self.string  # this returns the original code
        if self.items[1]:
            return f"{self.items[0]}:{self.items[1]}"
        return str(self.items[0])



# Copyright (c) 2023 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing pytest tests for the fparser2 CallBase class."""

import pytest
from fparser.two import Fortran2003
from fparser.two import utils


@pytest.mark.usefixtures("f2003_create")
def test_call_base_match():
    """Check that parent information is correctly set-up in the
    parse tree."""
    # Basic match is OK.
    assert utils.CallBase.match(
        Fortran2003.Procedure_Designator, Fortran2003.Actual_Arg_Spec_List, "ogg()"
    )
    # Trailing white space is ignored.
    assert utils.CallBase.match(
        Fortran2003.Procedure_Designator, Fortran2003.Actual_Arg_Spec_List, "ogg()  "
    )
    # String doesn't end with ')'
    assert not utils.CallBase.match(
        Fortran2003.Procedure_Designator, Fortran2003.Actual_Arg_Spec_List, "ogg"
    )
    # Missing opening '('
    assert not utils.CallBase.match(
        Fortran2003.Procedure_Designator, Fortran2003.Actual_Arg_Spec_List, "ogg)"
    )
    # Missing lhs.
    assert not utils.CallBase.match(
        Fortran2003.Procedure_Designator, Fortran2003.Actual_Arg_Spec_List, "(ogg)"
    )
    # Additional, matching parentheses are OK.
    assert utils.CallBase.match(
        Fortran2003.Procedure_Designator,
        Fortran2003.Actual_Arg_Spec_List,
        "ogg((nanny) )",
    )
    # upper_lhs makes no difference when lhs is matched with a class.
    assert utils.CallBase.match(
        Fortran2003.Procedure_Designator,
        Fortran2003.Actual_Arg_Spec_List,
        "ogg()",
        upper_lhs=True,
    )
    # upper_lhs is respected when lhs is matched with a str
    assert not utils.CallBase.match(
        "ogg", Fortran2003.Actual_Arg_Spec_List, "ogg()  ", upper_lhs=True
    )
    assert utils.CallBase.match(
        "OGG", Fortran2003.Actual_Arg_Spec_List, "ogg()  ", upper_lhs=True
    ) == ("OGG", None)
    # rhs can be matched using a str
    assert utils.CallBase.match(
        "ogg",
        "nanny",
        "ogg(nanny)",
    )
    assert not utils.CallBase.match(
        "ogg",
        "nanny",
        "ogg(granny)",
    )
    # require_rhs is respected
    assert utils.CallBase.match("ogg", "nanny", "ogg(nanny)", require_rhs=False) == (
        "ogg",
        "nanny",
    )
    assert utils.CallBase.match("ogg", "nanny", "ogg(nanny)", require_rhs=True) == (
        "ogg",
        "nanny",
    )
    assert not utils.CallBase.match("ogg", "nanny", "ogg()", require_rhs=True)


@pytest.mark.usefixtures("f2003_create")
def test_call_base_tostr():
    """Test the tostr() method of CallBase."""
    fref = Fortran2003.Function_Reference("gytha(ogg)")
    assert fref.tostr() == "gytha(ogg)"
    fref = Fortran2003.Function_Reference("weatherwax( )")
    assert fref.tostr() == "weatherwax()"



# Modified work Copyright (c) 2017-2024 Science and Technology
# Facilities Council.
# Original work Copyright (c) 1999-2008 Pearu Peterson
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Module containing py.test tests for the Fortran2003 OPEN statement (R904).

"""

import pytest

from fparser.two.Fortran2003 import Open_Stmt
from fparser.two.utils import NoMatchError


def test_open_stmt():
    """Tests that we correctly parse and re-generate the various forms
    of OPEN statement (R904)."""
    tcls = Open_Stmt
    obj = tcls("open(23, file='some_file.txt')")
    assert isinstance(obj, tcls)
    assert str(obj) == "OPEN(UNIT = 23, FILE = 'some_file.txt')"
    obj = tcls("open(unit=23, file='some_file.txt')")
    assert isinstance(obj, tcls)
    assert str(obj) == "OPEN(UNIT = 23, FILE = 'some_file.txt')"
    obj = tcls("open(unit=23, file='some_file.txt', convert=endianness)")
    assert isinstance(obj, tcls)
    assert str(obj) == "OPEN(UNIT = 23, FILE = 'some_file.txt', CONVERT = endianness)"
    obj = tcls(
        "OPEN (993,FILE=FNAMETAB,form='UNFORMATTED',convert=file_endian,IOSTAT=IERR,STATUS='OLD')"
    )
    assert isinstance(obj, tcls)
    assert (
        str(obj)
        == "OPEN(UNIT = 993, FILE = FNAMETAB, FORM = 'UNFORMATTED', CONVERT = file_endian, IOSTAT = IERR, STATUS = 'OLD')"
    )
    # Incorrect spelling of OPEN.
    with pytest.raises(NoMatchError) as err:
        tcls("opn(23, file='yada')")
    assert "Open_Stmt: 'opn(23," in str(err)



# Copyright (c) 2020 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2008 rule R510

deferred-coshape-spec is :

"""

import pytest
from fparser.two.Fortran2008 import Deferred_Coshape_Spec
from fparser.two import Fortran2003


@pytest.mark.usefixtures("f2008_create")
def test_deferred_coshape_spec():
    """Test parsing of deferred_coshape_spec."""
    obj = Deferred_Coshape_Spec(":")
    assert isinstance(obj, Deferred_Coshape_Spec), repr(obj)
    assert str(obj) == ":"


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize("attr", ["", " :", ": ", "  : ", "::", "5"])
def test_invalid_deferred_coshape_spec(attr):
    """Test that invalid deferred_coshape_spec raise exception."""
    with pytest.raises(Fortran2003.NoMatchError):
        _ = Deferred_Coshape_Spec(attr)



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Critical_Construct rule R810
"""

from fparser.two.Fortran2003 import Execution_Part_Construct
from fparser.two.Fortran2008.critical_stmt_r811 import Critical_Stmt
from fparser.two.Fortran2008.end_critical_stmt_r812 import End_Critical_Stmt
from fparser.two.utils import BlockBase


class Critical_Construct(BlockBase):
    """
    Fortran 2008 Rule 810.

    critical-construct is critical-stmt
                            block
                            end-critical-stmt

    TODO: Should disallow RETURN (C809) and CYCLE or EXIT to outside block (C811)
    """

    subclass_names = []
    use_names = ["Critical_Stmt", "Execution_Part_Construct", "End_Critical_Stmt"]

    @staticmethod
    def match(reader):
        """
        Attempt to match the supplied content with this Rule.

        :param reader: the fortran file reader containing the line(s)
                      of code that we are trying to match
        :type reader: :py:class:`fparser.common.readfortran.FortranFileReader` \
            | :py:class:`fparser.common.readfortran.FortranStringReader`

        :returns: instance of class that has matched or `None` if no match.
        :rtype: :py:class:`fparser.two.utils.BlockBase` | NoneType

        """
        return BlockBase.match(
            Critical_Stmt,
            [Execution_Part_Construct],
            End_Critical_Stmt,
            reader,
            match_names=True,  # C810
            strict_match_names=True,  # C810
        )



# -*- coding: utf-8 -*-
##############################################################################
# Copyright (c) 2017-2023 Science and Technology Facilities Council.
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##############################################################################
# Modified M.Hambley, UK Met Office
# Modified by J. Henrichs, Bureau of Meteorology
##############################################################################
"""
Test battery associated with fparser.sourceinfo package.
"""

import os
import tempfile
import pytest

from fparser.common.sourceinfo import (
    FortranFormat,
    get_source_info_str,
    get_source_info,
)

##############################################################################


def test_format_constructor_faults():
    """
    Tests that the constructor correctly rejects attempts to create an object
    with "None" arguments.
    """
    with pytest.raises(Exception):
        _unit_under_test = FortranFormat(True, None)

    with pytest.raises(Exception):
        _unit_under_test = FortranFormat(None, True)


##############################################################################
@pytest.fixture(
    scope="module",
    params=[
        (False, False, "fix", "Non-strict fixed format"),
        (False, True, "f77", "Strict fixed format"),
        (True, False, "free", "Non-strict free format"),
        (True, True, "pyf", "Strict free format"),
    ],
)
def pretty(request):
    """
    Returns all possible permutations of flags and their corresponding
    mode and descriptive strings.
    """
    return request.param


##############################################################################


def test_fortranformat_constructor(pretty):
    # pylint: disable=redefined-outer-name
    """
    Tests the constructor correctly sets up the object.
    """
    unit_under_test = FortranFormat(pretty[0], pretty[1])
    assert str(unit_under_test) == pretty[3]
    assert unit_under_test.is_free == pretty[0]
    assert unit_under_test.is_fixed == (not pretty[0])
    assert unit_under_test.is_strict == pretty[1]
    assert unit_under_test.is_f77 == (not pretty[0] and pretty[1])
    assert unit_under_test.is_fix == (not pretty[0] and not pretty[1])
    assert unit_under_test.is_pyf == (pretty[0] and pretty[1])
    assert unit_under_test.mode == pretty[2]
    assert not unit_under_test.f2py_enabled


@pytest.mark.parametrize(
    "permutations", [(False, False), (False, True), (True, False), (True, True)]
)
def test_fortranformat_equality(permutations, pretty):
    # pylint: disable=redefined-outer-name
    """
    Tests that the equality operator works as expected.
    """
    expected = (permutations[0] == pretty[0]) and (permutations[1] == pretty[1])
    unit_under_test = FortranFormat(permutations[0], permutations[1])
    candidate = FortranFormat(pretty[0], pretty[1])
    assert (unit_under_test == candidate) == expected


##############################################################################


def test_fortranformat_invalid():
    """
    Tests that the equality operator understands that it can't compare apples
    and oranges.
    """
    unit_under_test = FortranFormat(True, False)
    with pytest.raises(NotImplementedError):
        if unit_under_test == "oranges":
            raise Exception("That shouldn't have happened")


##############################################################################


@pytest.fixture(
    scope="module",
    params=[
        ("free", True, False),
        ("f77", False, True),
        ("fix", False, False),
        ("pyf", True, True),
    ],
)
def mode(request):
    """
    Returns all possible mode strings and their corresponding flags.
    """
    return request.param


##############################################################################


def test_fortranformat_from_mode(mode):
    # pylint: disable=redefined-outer-name
    """
    Tests that the object is correctly created by the from_mode function.
    """
    unit_under_test = FortranFormat.from_mode(mode[0])
    assert unit_under_test.mode == mode[0]
    assert unit_under_test.is_free == mode[1]
    assert unit_under_test.is_fixed == (not mode[1])
    assert unit_under_test.is_strict == mode[2]
    assert unit_under_test.is_f77 == (not mode[1] and mode[2])
    assert unit_under_test.is_fix == (not mode[1] and not mode[2])
    assert unit_under_test.is_pyf == (mode[1] and mode[2])
    assert str(unit_under_test.mode) == mode[0]
    assert not unit_under_test.f2py_enabled


def test_format_from_mode_bad():
    """
    Tests that an exception is thrown for an unrecognised mode string.
    """
    with pytest.raises(NotImplementedError):
        _unit_under_test = FortranFormat.from_mode("cheese")


##############################################################################
# Setting up a pytest fixture in this way is a mechanism for creating
# parameterised tests.
#
# Normally when a test includes a fixture in its argument list the
# corresponding function will be called and the result passed in to the test.
#
# When used like this the test will be called once for each value the fixture
# function returns. So in this case any test including "header" in its
# argument list will be run once with "header" equal to "! -*- fortran -*-",
# then with "header" equal to "! -*- f77 -*-" and so on.
#
# If a test includes multiple parameter fixtures it will be called for every
# permutation thus afforded.
#
@pytest.fixture(
    scope="module",
    params=[
        (None, FortranFormat(True, True)),
        ("! -*- fortran -*-", FortranFormat(False, True)),
        ("! -*- f77 -*-", FortranFormat(False, True)),
        ("! -*- f90 -*-", FortranFormat(True, False)),
        ("! -*- f03 -*-", FortranFormat(True, False)),
        ("! -*- f08 -*-", FortranFormat(True, False)),
        ("! -*- fix -*-", FortranFormat(False, False)),
        ("! -*- pyf -*-", FortranFormat(True, True)),
    ],
    name="header",
)
def header_fixture(request):
    """
    Returns parameters for header tests.
    """
    return request.param


##############################################################################

_FIXED_SOURCE = """      program main
      end program main
"""

_FREE_SOURCE = """program main
end program main
"""

_FIXED_WITH_CONTINUE = """      program main
          implicit none
          integer :: foo, &
                     bar
      end program main
"""

_FIXED_WITH_COMMENTS = """!     The program
      program main
c         Enforce explicit variable declaration
          implicit none
*         variables
          integer :: foo
      end program main
"""

# Tabs are not actually in the Fortran character set but fparser has handled
# them in the past even if it shouldn't have. We have to continue handling
# them for the time being until we work out why they were supported.
#
_INITIAL_TAB = "\tprogram main\n"
_MIDDLE_TAB = " \tprogram main\n"


# Another parameterised test fixture. See "header" above.
#
@pytest.fixture(
    scope="module",
    params=[
        (None, FortranFormat(False, False)),
        (_FIXED_SOURCE, FortranFormat(False, False)),
        (_FREE_SOURCE, FortranFormat(True, False)),
        (_FIXED_WITH_CONTINUE, FortranFormat(True, False)),
        (_FIXED_WITH_COMMENTS, FortranFormat(False, False)),
        (_INITIAL_TAB, FortranFormat(False, False)),
        (_MIDDLE_TAB, FortranFormat(True, False)),
    ],
)
def content(request):
    """
    Returns parameters for content tests.
    """
    return request.param


##############################################################################


def test_get_source_info_str(header, content):
    # pylint: disable=redefined-outer-name
    """
    Tests that source format is correctly identified when read from a string.
    """
    full_source = ""
    if header[0] is not None:
        full_source += header[0] + "\n"
    if content[0] is not None:
        full_source += content[0]

    source_info = get_source_info_str(full_source, ignore_encoding=False)
    if header[0]:
        assert source_info == header[1]
    else:  # No header
        assert source_info == content[1]


##############################################################################


# Another parameterised test fixture. See "header" above.
#
@pytest.fixture(
    scope="module",
    params=[
        (".f", None),
        (".f90", None),
        (".pyf", FortranFormat(True, True)),
        (".guff", None),
    ],
)
def extension(request):
    """
    Returns parameters for extension tests.
    """
    return request.param


##############################################################################


def test_get_source_info_filename(tmpdir, extension, header, content):
    # pylint: disable=redefined-outer-name
    """
    Tests that source format is correctly identified when read from a file.
    """
    full_source = ""
    if header[0] is not None:
        full_source += header[0] + "\n"
    if content[0] is not None:
        full_source += content[0]

    filename = f"{tmpdir}/out{extension[0]}"

    with open(filename, "w") as source_file:
        print(full_source, file=source_file)

    try:
        source_info = get_source_info(filename, ignore_encoding=False)
        if extension[1] is not None:
            assert source_info == extension[1]
        elif header[0] is not None:
            assert source_info == header[1]
        else:  # No header
            assert source_info == content[1]
    except Exception as ex:
        os.remove(filename)
        raise ex


##############################################################################


def test_get_source_info_file(extension, header, content):
    # pylint: disable=redefined-outer-name
    """
    Tests that source format is correctly identified when read from a file.
    """
    full_source = ""
    if header[0] is not None:
        full_source += header[0] + "\n"
    if content[0] is not None:
        full_source += content[0]

    with tempfile.TemporaryFile(mode="w+", suffix=extension[0]) as source_file:
        print(full_source, file=source_file)
        source_file.seek(0)

        source_info = get_source_info(source_file, ignore_encoding=False)
        if header[0] is not None:
            assert source_info == header[1]
        else:  # No header
            assert source_info == content[1]


def test_get_source_info_utf8():
    """
    Tests that Fortran code containing a unicode character can be read
    by the get_source_info method.

    """
    encoding = dict(encoding="UTF-8")
    with tempfile.NamedTemporaryFile(mode="w", **encoding) as tmp_file:
        content = '  ! A fortran comment with a unicode character "\u2014"'
        tmp_file.write(content)
        tmp_file.flush()

        source_info = get_source_info(tmp_file.name)
    assert source_info is not None


##############################################################################


def test_get_source_info_wrong():
    """
    Tests that get_source_info throws an exception if passed the wrong type
    of argument.
    """
    with pytest.raises(ValueError):
        _source_info = get_source_info(42)  # Obviously wrong

    with pytest.raises(ValueError):
        _source_info = get_source_info(["one"])  # Less obviously wrong


##############################################################################



# Copyright (c) 2018-2022 Science and Technology Facilities Council.
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing tests for the parser file"""

import pytest
from fparser.two.parser import ParserFactory
from fparser.common.readfortran import FortranStringReader
from fparser.two.utils import FortranSyntaxError, StmtBase
from fparser.two.symbol_table import SYMBOL_TABLES
from fparser.two import Fortran2003, Fortran2008


def test_parserfactory_std():
    """Test ParserFactory std argument options [none, f2003, f2008 and
    invalid]. Also tests that the SYMBOL_TABLES instance has been initialised
    correctly and that previous calls to the create method in the ParserFactory
    class do not affect current calls.

    """
    fstring = "submodule (x) y\n" "end\n"
    parser = ParserFactory().create()
    reader = FortranStringReader(fstring)
    with pytest.raises(FortranSyntaxError) as excinfo:
        _ = parser(reader)
    assert "at line 1\n>>>submodule (x) y\n" in str(excinfo.value)

    parser = ParserFactory().create(std="f2003")
    reader = FortranStringReader(fstring)
    with pytest.raises(FortranSyntaxError) as excinfo:
        _ = parser(reader)
    assert "at line 1\n>>>submodule (x) y\n" in str(excinfo.value)

    parser = ParserFactory().create(std="f2008")
    reader = FortranStringReader(fstring)
    ast = parser(reader)
    code = str(ast)
    assert "SUBMODULE (x) y\nEND" in code
    # Submodule_Stmt should now be included in the list of classes that define
    # scoping regions.
    assert "y" in SYMBOL_TABLES._symbol_tables

    # Repeat f2003 example to make sure that a previously valid (f2008)
    # match does not affect the current (f2003) invalid match.
    parser = ParserFactory().create(std="f2003")
    reader = FortranStringReader(fstring)
    with pytest.raises(FortranSyntaxError) as excinfo:
        _ = parser(reader)
    assert "at line 1\n>>>submodule (x) y\n" in str(excinfo.value)
    # The previous symbol table entries should have been removed when
    # creating the new parser.
    assert "y" not in SYMBOL_TABLES._symbol_tables

    with pytest.raises(ValueError) as excinfo:
        parser = ParserFactory().create(std="invalid")
        assert "is an invalid standard" in str(excinfo.value)


def _cmp_tree_types_rec(
    node1: Fortran2003.Program, node2: Fortran2003.Program, depth: int = 0
):
    """Helper function to recursively check for deepcopied programs

    :param node1: First AST tree to check
    :type node1: Fortran2003.Program
    :param node2: Second AST tree to check
    :type node2: Fortran2003.Program
    :param depth: Depth useful later on for debugging reasons,
        defaults to 0
    :type depth: int, optional
    """

    # Make sure that both trees are the same
    assert type(node1) is type(
        node2
    ), f"Nodes have different types: '{type(node1)}' and '{type(node2)}"

    if node1 is None:
        # Just return for None objects
        return

    if type(node1) is str:
        # WARNING: Different string objects with the same can have the same id.
        # Therefore, we can't compare with 'is' or with 'id(.) == id(.)'.
        # We can just compare the both strings have the same content.
        # See https://stackoverflow.com/questions/20753364/why-does-creating-multiple-objects-without-naming-them-result-in-them-having-the
        assert node1 == node2
        return

    else:
        # Make sure that we're working on a copy rather than the same object
        assert node1 is not node2, "Nodes refer to the same object"

    # Continue recursive traversal of ast
    for child1, child2 in zip(node1.children, node2.children):
        _cmp_tree_types_rec(child1, child2, depth + 1)


_f90_source_test = """
module andy
implicit none

   real :: apple = 1.0
   real, parameter :: pi = 3.14

contains
   subroutine sergi()
      print *, "Pi = ", pi
      print *,  "apple = ", apple
   end subroutine

end module andy


program awesome
    use andy
    implicit none

   real :: x
   integer :: i

   x = 2.2
   i = 7

   call sergi()

   print *, "apple pie: ", apple, pi
   print *, "i: ", i

end program awesome

"""


def test_deepcopy():
    """
    Test that we can deepcopy a parsed fparser tree.
    """

    parser = ParserFactory().create(std="f2008")
    reader = FortranStringReader(_f90_source_test)
    ast = parser(reader)

    import copy

    new_ast = copy.deepcopy(ast)

    _cmp_tree_types_rec(new_ast, ast)


def test_pickle():
    """
    Test that we can pickle and unpickle a parsed fparser tree.
    """

    parser = ParserFactory().create(std="f2008")
    reader = FortranStringReader(_f90_source_test)
    ast = parser(reader)

    import pickle

    s = pickle.dumps(ast)
    new_ast = pickle.loads(s)

    _cmp_tree_types_rec(new_ast, ast)



# Copyright (c) 2022, Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Test the setup performed for the fparser module.

"""

import os
import fparser


def test_fparser_get_version(monkeypatch):
    """Test the _get_version() utility routine. It doesn't make sense to
    actually check precisely which version it reports - just that it returns a
    suitable string."""
    ver1 = fparser._get_version()
    assert isinstance(ver1, str)
    assert "." in ver1

    def _broken_version(_name):
        """Broken routine with which to patch the `version` method."""
        raise fparser.PackageNotFoundError()

    from importlib import metadata

    monkeypatch.setattr(metadata, "version", _broken_version)
    ver2 = fparser._get_version()
    assert isinstance(ver2, str)
    assert "." in ver2


def test_fparser_logging_handler(tmpdir, caplog):
    """Test the custom error handler that is configured in the __init__.py
    file.  Invalid characters in an input file are skipped and logging
    messages are created.

    """
    content = "HELLO"
    invalid_content = "\xca".join(content)
    filepath = os.path.join(str(tmpdir), "tmp_in.f90")
    # Create the input file
    with open(filepath, "w", encoding="UTF-8") as tmp_file:
        tmp_file.write(invalid_content)
    # Specify encoding as 'ascii' to trigger errors for the non-ascii chars.
    with open(filepath, "r", errors="fparser-logging", encoding="ascii") as cfile:
        output = cfile.read()
    assert output == content
    for record in caplog.records:
        assert record.levelname != "CRITICAL"
    assert (
        "Skipped bad character in input file. Error returned was 'ascii' "
        "codec can't decode byte "
    ) in caplog.text

    assert (
        "Skipped bad character in input file. Error returned was 'ascii' "
        "codec can't decode byte 0xc3 in position 1: ordinal not in "
        "range(128)." in caplog.text
    )



# Copyright (c) 2017-2021 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing pytest tests for fparser2 base classes"""

import pytest
from fparser.common.readfortran import FortranStringReader
from fparser.two import Fortran2003
from fparser.two.symbol_table import SYMBOL_TABLES
from fparser.two.utils import NoMatchError, BlockBase, walk


@pytest.mark.usefixtures("f2003_create")
def test_keywordvaluebase_errors():
    """Unit tests for the KeywordValueBase class to check that it rejects
    invalid input"""
    from fparser.two.Fortran2003 import (
        KeywordValueBase,
        Io_Unit,
        Format,
        Char_Literal_Constant,
    )

    lhs_cls = "UNIT"
    rhs_cls = Io_Unit
    obj = KeywordValueBase.match(lhs_cls, rhs_cls, "  ", require_lhs=False)
    assert obj is None
    obj = KeywordValueBase.match(lhs_cls, rhs_cls, " = 36 ")
    assert obj is None
    _, obj = KeywordValueBase.match("FMT", Format, "'(A)'", require_lhs=False)
    assert isinstance(obj, Char_Literal_Constant)
    _, obj = KeywordValueBase.match("FMT", Format, "FMT='(A)'", require_lhs=False)
    assert isinstance(obj, Char_Literal_Constant)
    _, obj = KeywordValueBase.match(
        "FMT", Format, "FMT='(\"my_var =  \", (A))'", require_lhs=False
    )
    assert isinstance(obj, Char_Literal_Constant)
    lhs, obj = KeywordValueBase.match(
        "FMT", Format, "FMT='(\"my_var =  \", (A))'", require_lhs=True
    )
    assert lhs == "FMT"
    assert isinstance(obj, Char_Literal_Constant)
    # Try to trigger an error by specifying that a "LHS =" is required but
    # then omitting it while having the RHS contain a '=' character
    obj = KeywordValueBase.match(
        "FMT", Format, "'(\"my_var =  \", (A))'", require_lhs=True
    )
    assert obj is None
    # Check with a valid Fortran name on the lhs but not what the match
    # is requesting.
    with pytest.raises(NoMatchError) as err:
        _ = KeywordValueBase.match("FERMAT", Format, "FMT='(A)'", require_lhs=False)
    assert "FMT='(A)'" in str(err.value)


@pytest.mark.usefixtures("f2003_create")
def test_read_stmt_errors():
    """Unit tests for the Read class to ensure it rejects invalid
    inputs"""
    from fparser.two.Fortran2003 import Read_Stmt

    # Missing closing parenthesis
    obj = Read_Stmt.match("READ(unit=23")
    assert obj is None
    # Missing arguments
    obj = Read_Stmt.match("READ()")
    assert obj is None
    obj = Read_Stmt.match("READ")
    assert obj is None
    # Wrong argument type
    obj = Read_Stmt.match("READ a_var")
    assert obj is None
    obj = Read_Stmt.match("READ 13")
    assert obj is None
    # Missing comma
    obj = Read_Stmt.match("READ * a_var")
    assert obj is None
    # Missing value/variable after comma
    obj = Read_Stmt.match("READ 13, ")
    assert obj is None


@pytest.mark.usefixtures("f2003_create")
def test_io_ctrl_spec_list_errors():
    """Unit tests for the Io_Control_Spec_List class to ensure it
    rejects invalid input"""
    from fparser.two.Fortran2003 import Io_Control_Spec_List

    # Positional arg following named arg
    obj = Io_Control_Spec_List.match("unit=23, namvar")
    assert obj is None


@pytest.mark.usefixtures("f2003_create")
def test_io_ctrl_spec_errors():
    """Unit tests for the Io_Control_Spec class to ensure it
    rejects invalid input"""
    from fparser.two.Fortran2003 import Io_Control_Spec

    # An argument with a name that is not valid within an IO control
    # description
    obj = Io_Control_Spec.match("not_unit=23")
    assert obj is None


@pytest.mark.usefixtures("f2003_create")
def test_blockbase_tofortran_non_ascii():
    """Check that the tofortran() method works when we have a program
    containing non-ascii characters within a sub-class of BlockBase. We
    use a Case Construct for this purpose."""
    from fparser.two.Fortran2003 import Program, Case_Construct

    code = (
        "program my_test\n"
        "! A comment outside the select block\n"
        "SELECT CASE(iflag)\n"
        "CASE(  30  )\n"
        "  IF(lwp) WRITE(*,*) ' for e1=1\xb0'\n"
        "END SELECT\n"
        "end program\n"
    )
    reader = FortranStringReader(code, ignore_comments=False)
    obj = Program(reader)
    bbase = walk(obj.content, Case_Construct)[0]
    # Explicitly call tofortran() on the BlockBase class.
    out_str = BlockBase.tofortran(bbase)
    assert "for e1=1" in out_str



# This Python file uses the following encoding: utf-8
# Copyright (c) 2020-2022 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""pytest module for the Fortran2003 Where Construct - R744. Note that
this requires re-implementing and extending to fully cover the
language specification (#232)."""

import pytest
from fparser.api import get_reader
from fparser.common.readfortran import FortranStringReader
from fparser.two.Fortran2003 import Where_Construct
from fparser.two.utils import FortranSyntaxError


@pytest.mark.usefixtures("f2003_create")
def test_where_construct():
    """Tests for the WHERE construct, R744."""
    tcls = Where_Construct
    obj = tcls(get_reader("""\
    where (pressure <= 1.0)
    pressure = pressure + inc_pressure
    temp = temp - 5.0
    elsewhere
    raining = .true.
    end where
"""))
    assert isinstance(obj, tcls), repr(obj)
    assert (
        str(obj) == "WHERE (pressure <= 1.0)\n  "
        "pressure = pressure + inc_pressure\n  "
        "temp = temp - 5.0\n"
        "ELSEWHERE\n  raining = .TRUE.\nEND WHERE"
    )

    obj = tcls(get_reader("""\
    where (cond1)
    else    where (cond2)
    end where
"""))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "WHERE (cond1)\nELSEWHERE(cond2)\nEND WHERE"

    obj = tcls(get_reader("""\
    n:where (cond1)
    elsewhere (cond2) n
    else   where n
    end where n
"""))
    assert isinstance(obj, tcls), repr(obj)
    assert (
        str(obj) == "n:WHERE (cond1)\nELSEWHERE(cond2) n\n" "ELSEWHERE n\nEND WHERE n"
    )

    obj = tcls(get_reader("""\
    n:where (cond1)
    else where (cond2) n
    else where n
    end where n
"""))
    assert isinstance(obj, tcls), repr(obj)
    assert (
        str(obj) == "n:WHERE (cond1)\nELSEWHERE(cond2) n\nELSEWHERE n\n" "END WHERE n"
    )

    obj = tcls(get_reader("""\
    n:where (me(:)=="hello")
    else where (me(:)=="goodbye") n
    else where n
    end where n
"""))
    assert (
        str(obj) == 'n:WHERE (me(:) == "hello")\nELSEWHERE(me(:) == "goodbye") n\n'
        "ELSEWHERE n\n"
        "END WHERE n"
    )


@pytest.mark.usefixtures("f2003_create")
def test_where_tofortran_non_ascii():
    """Check that the tofortran() method works when a WHERE construct
    contains a character string with non-ascii characters."""
    code = "WHERE(iflag)\n" "  msg_list = ' for e1=1\xb0'\n" "END WHERE\n"
    reader = FortranStringReader(code, ignore_comments=False)
    obj = Where_Construct(reader)
    out_str = str(obj)
    assert "for e1=1" in out_str


def test_where_construct_wrong_name(f2003_create, fake_symbol_table):
    """Check named 'where' construct has correct start/end name"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Where_Construct(get_reader("""\
            name: where (expr)
                a = 1
            end where wrong"""))
    assert exc_info.value.args[0].endswith("Expecting name 'name', got 'wrong'")


def test_where_construct_missing_start_name(f2003_create, fake_symbol_table):
    """Check named 'where' construct has correct start/end name"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Where_Construct(get_reader("""\
            where (expr)
                a = 1
            end where name"""))
    assert exc_info.value.args[0].endswith(
        "Name 'name' has no corresponding starting name"
    )


def test_where_construct_missing_end_name(f2003_create, fake_symbol_table):
    """Check named 'where' construct has correct start/end name"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Where_Construct(get_reader("""\
            name: where (expr)
                a = 1
            end where"""))
    assert exc_info.value.args[0].endswith("Expecting name 'name' but none given")


def test_where_construct_else_wrong_end_name(f2003_create, fake_symbol_table):
    """Check named 'where' construct has correct start/end name"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Where_Construct(get_reader("""\
            name: where (expr)
                a = 1
            elsewhere
                a = 2
            end where wrong"""))
    assert exc_info.value.args[0].endswith("Expecting name 'name', got 'wrong'")


def test_where_construct_else_wrong_name(f2003_create, fake_symbol_table):
    """Check named 'where' construct has correct start/end name"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Where_Construct(get_reader("""\
            name: where (expr)
                a = 1
            elsewhere wrong
                a = 2
            end where name"""))
    assert exc_info.value.args[0].endswith("Expecting name 'name', got 'wrong'")


def test_where_construct_else_where_wrong_name(f2003_create, fake_symbol_table):
    """Check named 'where' construct has correct start/end name"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Where_Construct(get_reader("""\
            name: where (expr)
                a = 1
            elsewhere (other_expr) wrong
                a = 2
            end where name"""))
    assert exc_info.value.args[0].endswith("Expecting name 'name', got 'wrong'")



# Copyright (c) 2022 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R1111 : This file tests the support for the
Rename class.

"""

from fparser.two.Fortran2003 import Rename


def test_rename():
    """
    Check basic functionality of the Rename class.
    """
    tcls = Rename
    obj = tcls("a=>b")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "a => b"
    assert obj.children[0] is None

    obj = tcls("operator(.foo.)=>operator(.bar.)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "OPERATOR(.FOO.) => OPERATOR(.BAR.)"
    assert obj.children[0] == "OPERATOR"

    obj = tcls("operator_1=>operator_2")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "operator_1 => operator_2"
    assert obj.children[0] is None


def test_rename_nomatch():
    """
    Check that various syntax errors result in no match.
    """
    assert Rename.match("operator(.foo.) = operator(.bar.)") is None
    assert Rename.match("operator(.foo.) => ") is None
    assert Rename.match("operator(.foo.) => operator_1") is None
    assert Rename.match("operator() => operator(.bar.)") is None



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Attr_Spec rule R502
"""

from fparser.two import pattern_tools as pattern
from fparser.two.Fortran2003 import Attr_Spec as Attr_Spec_2003
from fparser.two.utils import STRINGBase


class Attr_Spec(Attr_Spec_2003):  # R502
    """
    Fortran 2008 rule R502.

    .. code-block:: fortran

        attr-spec is access-spec
                     or ALLOCATABLE
                     or ASYNCHRONOUS
                     or CODIMENSION lbracket coarray-spec rbracket
                     or CONTIGUOUS
                     or DIMENSION ( array-spec )
                     or EXTERNAL
                     or INTENT ( intent-spec )
                     or INTRINSIC
                     or language-binding-spec
                     or OPTIONAL
                     or PARAMETER
                     or POINTER
                     or PROTECTED
                     or SAVE
                     or TARGET
                     or VALUE
                     or VOLATILE

    In the spec above, lbracket and rbracket are left and right square
    brackets `[]` but not printed explicitly to avoid misinterpretation
    as optional parts.

    This rule adds CODIMENSION and CONTIGUOUS attributes to Fortran2003's R503.

    """

    subclass_names = Attr_Spec_2003.subclass_names[:]
    subclass_names.append("Codimension_Attr_Spec")
    use_names = []

    @staticmethod
    def match(string):
        """
        Implements the matching for attributes of types.

        :param str string: the string to match as attribute.

        :return: `None` if there is no match, otherwise a 1-tuple \
                 containing the matched string.
        :rtype: `NoneType` or (`str`,)

        """
        return STRINGBase.match(pattern.abs_attr_spec_f08, string)



# Copyright (c) 2021 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R727 : This file tests the support for a
Fortran integer expression.

"""

import pytest
from fparser.two.Fortran2003 import Int_Expr, Int_Literal_Constant, Level_2_Expr
from fparser.two.utils import NoMatchError


@pytest.mark.usefixtures("f2003_create")
def test_simple_case():
    """Test that a simple expression gives the expected result."""
    result = Int_Expr("1")
    assert isinstance(result, Int_Literal_Constant)
    assert str(result) == "1"
    assert repr(result) == "Int_Literal_Constant('1', None)"


@pytest.mark.usefixtures("f2003_create")
def test_complicated_case():
    """Test that a more complicated expression gives the expected
    result.

    """
    result = Int_Expr("a*2+array(b)-w")
    assert isinstance(result, Level_2_Expr)
    assert str(result) == ("a * 2 + array(b) - w")
    assert repr(result).replace("u'", "'") == (
        "Level_2_Expr(Level_2_Expr(Add_Operand(Name('a'), '*', "
        "Int_Literal_Constant('2', None)), '+', Part_Ref(Name('array'), "
        "Section_Subscript_List(',', (Name('b'),)))), '-', Name('w'))"
    )


@pytest.mark.parametrize(
    "string", [".true.", "b'1010'", "o'7070'", "h'f0f0'", "1.0", "(1.0,1.0)", "'hello'"]
)
@pytest.mark.usefixtures("f2003_create")
def test_c708(string):
    """Check that invalid literal constants do not match. Note, there are
    many other cases that are not currently checked in fparser.

    In theory -1.0 should become a Signed_Real_Literal_Constant and
    then fail to match. However the "-" is treated as a unary
    expression so this never happens.

    """
    with pytest.raises(NoMatchError):
        _ = Int_Expr(string)



# Copyright (c) 2022-2024 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2008 rule R904

   open-stmt is OPEN ( connect-spec-list )

   R905 connect-spec is [ UNIT = ] file-unit-number
                    or ACCESS = scalar-default-char-expr
                    or ACTION = scalar-default-char-expr
                    or ASYNCHRONOUS = scalar-default-char-expr
                    or BLANK = scalar-default-char-expr
                    [ or CONVERT = scalar-default-char-expr ]
                    or DECIMAL = scalar-default-char-expr
                    or DELIM = scalar-default-char-expr
                    or ENCODING = scalar-default-char-expr
                    or ERR = label
                    or FILE = file-name-expr
                    or FORM = scalar-default-char-expr
                    or IOMSG = iomsg-variable
                    or IOSTAT = scalar-int-variable
                    or NEWUNIT = scalar-int-variable
                    or PAD = scalar-default-char-expr
                    or POSITION = scalar-default-char-expr
                    or RECL = scalar-int-expr
                    or ROUND = scalar-default-char-expr
                    or SIGN = scalar-default-char-expr
                    or STATUS = scalar-default-char-expr

Note that the CONVERT argument is non-standard but supported by several major
compilers.

R906 file-name-expr is scalar-default-char-expr
R907 iomsg-variable is scalar-default-char-variable

C903 No specifier shall appear more than once in a given connect-spec-list.
C904 (R904) If the NEWUNIT= specifier does not appear, a file-unit-number
     shall be specified; if the optional characters UNIT= are omitted, the
     file-unit-number shall be the first item in the connect-spec-list.
C905 (R904) The label used in the ERR= specifier shall be the statement label
     of a branch target statement that appears in the same inclusive scope as
     the OPEN statement.
C906 (R904) If a NEWUNIT= specifier appears, a file-unit-number shall not
     appear.

"""

import pytest
from fparser.api import get_reader
from fparser.two import Fortran2008
from fparser.two.utils import NoMatchError, walk


@pytest.mark.parametrize(
    "open_args",
    [
        "unit=23, file='hello'",
        "23, file='hello', err=36",
        "23, file='hello', status='OLD'",
        "unit=23, file='hello', action='read'",
        "unit=23, file='hello', convert=endianess, action='read'",
    ],
)
@pytest.mark.usefixtures("f2008_parser")
def test_open_f2003_args(open_args):
    """Check that the Fortran2008 version of Open_Stmt still supports the
    various arguments defined in 2003."""
    obj = Fortran2008.Open_Stmt(f"open({open_args})")
    assert isinstance(obj, Fortran2008.Open_Stmt)


def test_open_newunit(f2008_parser):
    """Test that NEWUNIT is a valid argument to OPEN."""
    tree = f2008_parser(get_reader("""\
subroutine myopen( unit, file )
  integer, intent(out):: unit
  character(len=*), intent(in):: file
  open( newunit=unit, file=file )
endsubroutine myopen
    """))
    open_stmts = walk(tree, Fortran2008.Open_Stmt)
    assert open_stmts
    assert len(open_stmts[0].children) == 2
    assert str(open_stmts[0]) == "OPEN(NEWUNIT = unit, FILE = file)"


@pytest.mark.usefixtures("f2008_parser")
def test_constraint_903():
    """Check that Constraint 903 (no specifier shall appear more than once)
    is applied."""
    with pytest.raises(NoMatchError):
        Fortran2008.Open_Stmt("open(23, unit=24, file='hello')")
    with pytest.raises(NoMatchError):
        Fortran2008.Open_Stmt("open(23, file='hello', file='another')")


@pytest.mark.usefixtures("f2008_parser")
def test_constraint_904():
    """Check that a unit number is specified. We cannot currently check that
    if no UNIT= appears that it is the first argument to open() because the
    Connect_Spec.match() method always adds a UNIT= if one is missing."""
    with pytest.raises(NoMatchError):
        Fortran2008.Open_Stmt("open(file='hello')")


@pytest.mark.usefixtures("f2008_parser")
def test_constraint_906():
    """Check that Constraint 906 is applied (no unit number may appear if
    NEWUNIT is specified)."""
    with pytest.raises(NoMatchError):
        Fortran2008.Open_Stmt("open(23, newunit=10, file='hello')")
    with pytest.raises(NoMatchError):
        Fortran2008.Open_Stmt("open(unit=23, newunit=10, file='hello')")


@pytest.mark.usefixtures("f2008_parser")
def test_open_invalid_arg():
    """Check that there is no match if an invalid argument is supplied."""
    with pytest.raises(NoMatchError) as err:
        Fortran2008.Open_Stmt("open(newunit=10, file='hello', andy='yes')")
    assert "andy='yes'" in str(err)



#!/usr/bin/env python
# Copyright (c) 2022 Science and Technology Facilities Council
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Generates a large Fortran program in memory and then measures how long
it takes fparser2 to parse it. This is based on the benchmark suggested
by Ondřej Čertík via Ioannis Nikiteas.

"""

from time import perf_counter

from fparser.common.sourceinfo import FortranFormat
from fparser.common.readfortran import FortranStringReader
from fparser.two.parser import ParserFactory


def gen_sub(num: int):
    """
    Constructs a Fortran subroutine named g<num>.

    :param num: the number of the subroutine (used to name it).

    :returns: Fortran subroutine.
    :rtype: str

    """
    sub = f"""subroutine g{num}(x)
    integer, intent(inout) :: x
    integer :: i
    x = 0
    do i = {num}, {num+9}
        x = x+i
    end do
end subroutine

"""
    return sub


def create_bench(num_routines: int):
    """
    Creates the Fortran benchmark code.

    :param num_routines: the number of subroutines to create.

    :returns: benchmark Fortran code.
    :rtype: str

    """
    code = ["program bench3", "implicit none", "integer :: c", "c = 0"]
    for i in range(1, num_routines + 1):
        code.append(f"call g{i}(c)")

    code.append("print *, c")
    code.append("contains")

    for i in range(1, num_routines + 1):
        code.append(gen_sub(i))
    code.append("end program")

    return "\n".join(code)


def runner(num_routines: int):
    """
    Entry point for running the benchmark.

    :param num_routines: the number of subroutines to create in the \
                         Fortran benchmark.

    :raises ValueError: if num_routines < 1.

    """
    if num_routines < 1:
        raise ValueError(
            f"Number of routines to create must be a positive, "
            f"non-zero integer but got: {num_routines}"
        )

    print(f"Constructing benchmark code with {num_routines} subroutines...")
    code = create_bench(num_routines)
    reader = FortranStringReader(code)
    # Ensure the reader uses free format.
    reader.set_format(FortranFormat(True, True))

    fparser = ParserFactory().create()

    print("Parsing benchmark code...")
    tstart = perf_counter()
    _ = fparser(reader)
    tstop = perf_counter()

    print(f"Time taken for parse = {tstop - tstart:.2f}s")


if __name__ == "__main__":
    runner(10000)  # pragma: no cover



# -----------------------------------------------------------------------------
# Copyright (c) 2021-2023 Science and Technology Facilities Council.
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""Module containing tests for the symbol-table functionality
of fparser2."""

import pytest
from fparser.api import get_reader
from fparser.two import Fortran2003
from fparser.two.parser import ParserFactory
from fparser.two.symbol_table import SymbolTable, SYMBOL_TABLES, SymbolTableError


def test_basic_table():
    """Check the basic functionality of a symbol table."""
    table = SymbolTable("BAsic")
    # Name of table is not case sensitive
    assert table.name == "basic"
    assert table.parent is None
    assert table.children == []
    assert table.node is None
    # Consistency checking is disabled by default
    assert table._checking_enabled is False
    with pytest.raises(KeyError) as err:
        table.lookup("missing")
    assert "Failed to find symbol named 'missing'" in str(err.value)
    # Add a symbol and check that its naming is not case sensitive
    table.add_data_symbol("Var", "integer")
    sym = table.lookup("var")
    assert sym.name == "var"
    assert table.lookup("VAR") is sym
    # Check that we can enable consistency checking
    table2 = SymbolTable("table2", checking_enabled=True)
    assert table2._checking_enabled is True
    # Check that we can supply an associated parse tree node.
    with pytest.raises(TypeError) as err:
        SymbolTable("table3", node="oops")
    assert (
        "The 'node' argument to the SymbolTable constructor must be a "
        "valid parse tree node (instance of utils.Base) but got 'str'" in str(err.value)
    )
    table3 = SymbolTable("table3", node=Fortran2003.Return_Stmt("return"))
    assert isinstance(table3.node, Fortran2003.Return_Stmt)


def test_add_data_symbol():
    """Test that the add_data_symbol() method behaves as expected when
    validation is enabled."""
    table = SymbolTable("basic", checking_enabled=True)
    table.add_data_symbol("var", "integer")
    sym = table.lookup("var")
    assert sym.primitive_type == "integer"
    with pytest.raises(SymbolTableError) as err:
        table.add_data_symbol("var", "real")
    assert (
        "Symbol table already contains a symbol for a variable with name "
        "'var'" in str(err.value)
    )
    with pytest.raises(TypeError) as err:
        table.add_data_symbol(table, "real")
    assert "name of the symbol must be a str but got 'SymbolTable'" in str(err.value)
    with pytest.raises(TypeError) as err:
        table.add_data_symbol("var2", table)
    assert (
        "primitive type of the symbol must be specified as a str but got "
        "'SymbolTable'" in str(err.value)
    )
    # Check a clash with a USE statement - both the module name and the
    # name of imported variables
    table.add_use_symbols("mod1", only_list=[("var3", None)])
    with pytest.raises(SymbolTableError) as err:
        table.add_data_symbol("mod1", "real")
    assert "table already contains a use of a module with name 'mod1'" in str(err.value)
    with pytest.raises(SymbolTableError) as err:
        table.add_data_symbol("var3", "real")
    assert (
        "table already contains a use of a symbol named 'var3' from "
        "module 'mod1'" in str(err.value)
    )


def test_add_data_symbols_no_checks():
    """Check that we can disable the checks in the
    add_data_symbol() method."""
    table = SymbolTable("basic", checking_enabled=False)
    table.add_data_symbol("var", "integer")
    table.add_data_symbol("var", "real")
    sym = table.lookup("var")
    assert sym.primitive_type == "real"
    table.add_use_symbols("mod1", [("var3", None)])
    table.add_data_symbol("mod1", "real")
    table.add_use_symbols("mod2", [("var3", None)])
    table.add_data_symbol("var3", "real")
    assert table.lookup("var3").primitive_type == "real"


def test_add_use_symbols():
    """Test that the add_use_symbols() method behaves as expected."""
    table = SymbolTable("basic")
    # A use without an 'only' clause
    table.add_use_symbols("mod1")
    assert table._modules["mod1"].only_list is None
    assert table._modules["mod1"].rename_list is None
    assert table._modules["mod1"].wildcard_import is True
    # Fortran permits other use statements for the same module
    table.add_use_symbols("mod1", only_list=[("var", None)])
    # Since we already have a wildcard import that should remain true while
    # we now also capture those symbols that are explicitly imported.
    assert table._modules["mod1"].only_list == ["var"]
    assert table._modules["mod1"].wildcard_import
    table.add_use_symbols("mod2", only_list=[("iVar", None)])
    assert table._modules["mod2"].only_list == ["ivar"]
    table.add_use_symbols("mod2", only_list=[("jvar", None)])
    assert sorted(table._modules["mod2"].only_list) == ["ivar", "jvar"]


def test_add_use_symbols_errors():
    """Test the various checks on the supplied parameters to
    add_use_symbols()."""
    table = SymbolTable("basic")
    with pytest.raises(TypeError) as err:
        table.add_use_symbols(table)
    assert "name of the module must be a str but got 'SymbolTable'" in str(err.value)
    with pytest.raises(TypeError) as err:
        table.add_use_symbols("mod3", only_list="hello")
    assert "If present, the only_list must be a list but got 'str'" in str(err.value)
    with pytest.raises(TypeError) as err:
        table.add_use_symbols("mod3", only_list=[("hello", None, None)])
    assert (
        "If present, the only_list must be a list of 2-tuples but got: "
        "[('hello', None, None)]" in str(err.value)
    )


def test_str_method():
    """Test the str property of the SymbolTable class."""
    table = SymbolTable("basic")
    assert "Symbol Table 'basic'\nSymbols:\nUsed modules:\n" in str(table)
    table.add_data_symbol("var", "integer")
    assert "Symbol Table 'basic'\nSymbols:\nvar\nUsed modules:\n" in str(table)
    table.add_use_symbols("some_mod")
    assert "Symbol Table 'basic'\nSymbols:\nvar\nUsed modules:\nsome_mod\n" in str(
        table
    )


def test_del_child():
    """Checks for the del_child method."""
    table = SymbolTable("BASIC")
    inner_table = SymbolTable("func1", parent=table)
    table.add_child(inner_table)
    with pytest.raises(KeyError) as err:
        table.del_child("missing")
    assert "Symbol table 'basic' does not contain a table named 'missing'" in str(
        err.value
    )
    table.del_child("func1")
    assert table.children == []


def test_parent_child():
    """Test the parent/child-related properties."""
    table = SymbolTable("BASIC")
    with pytest.raises(TypeError) as err:
        table.add_child("wrong")
    assert "Expected a SymbolTable instance but got 'str'" in str(err.value)
    inner_table = SymbolTable("func1", parent=table)
    table.add_child(inner_table)
    assert table.children == [inner_table]
    assert inner_table.parent is table
    with pytest.raises(TypeError) as err:
        inner_table.parent = "wrong"
    assert (
        "Unless it is None, the parent of a SymbolTable must also be a "
        "SymbolTable but got 'str'" in str(err.value)
    )


def test_root_property():
    """Test the `root` property of the SymbolTable."""
    table = SymbolTable("BASIC")
    inner_table = SymbolTable("func1", parent=table)
    table.add_child(inner_table)
    inner_inner_table = SymbolTable("func2", parent=inner_table)
    assert inner_inner_table.root is table
    assert inner_table.root is table
    assert table.root is table


def test_module_use(f2003_parser):
    """Check that a USE of a module is captured in the symbol table."""
    _ = f2003_parser(get_reader("""\
PROGRAM a_prog
  use some_mod
END PROGRAM a_prog
    """))
    tables = SYMBOL_TABLES
    table = tables.lookup("a_prog")
    assert isinstance(table, SymbolTable)
    assert table.parent is None
    assert "some_mod" in table._modules


def test_module_use_with_only(f2003_parser):
    """Check that USE statements with an ONLY: clause are correctly captured
    in the symbol table."""
    _ = f2003_parser(get_reader("""\
PROGRAM a_prog
  use some_mod, only:
  use mod2, only: this_one, that_one
END PROGRAM a_prog
    """))
    tables = SYMBOL_TABLES
    table = tables.lookup("a_prog")
    assert isinstance(table, SymbolTable)
    assert table.parent is None
    assert "some_mod" in table._modules
    assert table._modules["some_mod"].only_list == []
    assert "mod2" in table._modules
    assert sorted(table._modules["mod2"].only_list) == ["that_one", "this_one"]
    sym = table.lookup("this_one")
    assert sym.name == "this_one"
    assert sym.primitive_type == "unknown"
    sym = table.lookup("that_one")
    assert sym.name == "that_one"
    assert sym.primitive_type == "unknown"


def test_module_use_with_rename(f2003_parser):
    """Check that USE statements with renamed imported symbols are correctly
    captured in the symbol table and do not clash."""
    _ = f2003_parser(get_reader("""\
PROGRAM a_prog
  use mod2, only: this_one => that_one
  use mod3, local => other
  integer :: that_one
  logical :: other
END PROGRAM a_prog
    """))
    tables = SYMBOL_TABLES
    table = tables.lookup("a_prog")
    assert isinstance(table, SymbolTable)
    assert table.parent is None
    assert "mod2" in table._modules
    mod2 = table._modules["mod2"]
    assert mod2.only_list == ["this_one"]
    assert mod2.get_declared_name("this_one") == "that_one"
    sym = table.lookup("that_one")
    assert sym.primitive_type == "integer"
    mod3 = table._modules["mod3"]
    assert mod3.rename_list == ["local"]
    assert mod3.get_declared_name("local") == "other"
    assert table.lookup("other").primitive_type == "logical"


def test_wildcard_module_search(f2003_parser):
    """Test the wildcard_imports method of the SymbolTable."""
    _ = f2003_parser(get_reader("""\
module my_mod
  use other_mod, only: b
  use medium_mod
  use some_mod
  real :: a
contains
  subroutine sub
    use big_mod
    use medium_mod, only: c
    use pointless_mod, only:
  end subroutine sub
end module my_mod
    """))
    # Module symbol table should have two modules listed with wildcard imports.
    mod_table = SYMBOL_TABLES.lookup("my_mod")
    assert mod_table.wildcard_imports == ["medium_mod", "some_mod"]
    # Move down to the subroutine and check that we recurse upwards.
    sub_table = mod_table.children[0]
    assert sub_table.wildcard_imports == ["big_mod", "medium_mod", "some_mod"]
    # Repeat for a program containing a subroutine containing a function.
    _ = f2003_parser(get_reader("""\
program my_prog
  use pointless_mod, only: dee
  use medium_mod
  use no_really_mod
  real :: a
  call sub()
contains
  subroutine sub
    use big_mod
    use no_really_mod, only: avon
    use medium_mod, only: c
    use pointless_mod, only:
    write(*,*) "yes", my_fn()
  contains
    integer function my_fn()
      use no_really_mod, only: afon
      use tiny_mod
      my_fn = 1 + afon
    end function my_fn
  end subroutine sub
end program my_prog
    """))
    prog_table = SYMBOL_TABLES.lookup("my_prog")
    assert prog_table.wildcard_imports == ["medium_mod", "no_really_mod"]
    sub_table = prog_table.children[0]
    assert sub_table.wildcard_imports == ["big_mod", "medium_mod", "no_really_mod"]
    fn_table = sub_table.children[0]
    assert fn_table.wildcard_imports == [
        "big_mod",
        "medium_mod",
        "no_really_mod",
        "tiny_mod",
    ]


def test_module_definition(f2003_parser):
    """Check that a SymbolTable is created for a module and populated with
    the symbols it defines."""
    _ = f2003_parser(get_reader("""\
module my_mod
  use some_mod
  real :: a
end module my_mod
    """))
    tables = SYMBOL_TABLES
    assert list(tables._symbol_tables.keys()) == ["my_mod"]
    table = tables.lookup("my_mod")
    assert isinstance(table, SymbolTable)
    assert isinstance(table.node, Fortran2003.Module_Stmt)
    assert "some_mod" in table._modules
    assert "a" in table._data_symbols
    sym = table.lookup("a")
    assert sym.name == "a"
    assert sym.primitive_type == "real"


def test_routine_in_module(f2003_parser):
    """Check that we get two, nested symbol tables when a module contains
    a subroutine."""
    _ = f2003_parser(get_reader("""\
module my_mod
  use some_mod
  real :: a
contains
  subroutine my_sub()
  end subroutine my_sub
end module my_mod
    """))
    tables = SYMBOL_TABLES
    assert list(tables._symbol_tables.keys()) == ["my_mod"]
    table = tables.lookup("my_mod")
    assert isinstance(table.node, Fortran2003.Module_Stmt)
    assert len(table.children) == 1
    assert table.children[0].name == "my_sub"
    assert table.children[0].parent is table
    # Check that the search for a symbol moves up to the parent scope
    sym = table.children[0].lookup("a")
    assert sym.name == "a"
    assert sym.primitive_type == "real"


def test_routine_in_prog(f2003_parser):
    """Check that we get two, nested symbol tables when a program contains
    a subroutine."""
    _ = f2003_parser(get_reader("""\
program my_prog
  use some_mod
  real :: a
contains
  subroutine my_sub()
    real :: b
  end subroutine my_sub
end program my_prog
    """))
    tables = SYMBOL_TABLES
    assert list(tables._symbol_tables.keys()) == ["my_prog"]
    table = SYMBOL_TABLES.lookup("my_prog")
    assert len(table.children) == 1
    assert isinstance(table.node, Fortran2003.Program_Stmt)
    assert table.children[0].name == "my_sub"
    assert isinstance(table.children[0].node, Fortran2003.Subroutine_Stmt)
    assert table.children[0]._data_symbols["b"].name == "b"
    assert table.children[0].parent is table


def test_all_symbols_resolved(f2003_parser):
    """Tests for the all_symbols_resolved() method."""
    code = """\
program my_prog
  use some_mod
  real :: a
contains
  subroutine my_sub()
    real :: b
  end subroutine my_sub
end program my_prog
    """
    _ = f2003_parser(get_reader(code))
    table = SYMBOL_TABLES.lookup("my_prog").children[0]
    assert table.all_symbols_resolved is False
    new_code = code.replace("some_mod\n", "some_mod, only: igor\n")
    SYMBOL_TABLES.clear()
    _ = f2003_parser(get_reader(new_code))
    table = SYMBOL_TABLES.lookup("my_prog").children[0]
    assert table.all_symbols_resolved is True


def test_all_symbols_resolved_submodule():
    """Tests for the all_symbols_resolved() method returns False when the table
    is within a Fortran2008 submodule (since a submodule has access to the
    scope of its parent module)."""
    code = """\
submodule (other_mod) my_mod
  real :: a
contains
  subroutine my_sub()
    real :: b
  end subroutine my_sub
end submodule my_mod
    """
    _ = ParserFactory().create(std="f2008")(get_reader(code))
    table = SYMBOL_TABLES.lookup("my_mod").children[0]
    assert table.all_symbols_resolved is False
